(* Pack single files into zip or 7z

     Dr. J. Rathlev, D-24222 Schwentinental (kontakt(a)rathlev-home.de)

   Acknowledgements:
     ZLibEx and ZLib routines from http://www.base2ti.com
     AES functions from http://fp.gladman.plus.com/index.htm

   The contents of this file may be used under the terms of the
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Vers. 1 - March 2011
   Vers. 2 - Feb. 2021: zip,7z (Delphi 10)
   last modified:
   *)

unit SevenZipMain;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes,
  Vcl.Graphics, Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.Buttons, Vcl.StdCtrls,
  HListBox, Vcl.ComCtrls, Vcl.ExtCtrls;

const
  CopRgt = ' 2011-2021 Dr. J. Rathlev, 24222 Schwentinental';
  EmailAdr = 'kontakt(a)rathlev-home.de';

  Name7Zip   = '7z.exe';
  Path7Zip   = '7-Zip\';

type
  TArcType = (atZip,at7z);
  TArcProp = record
    TypeId,Ext : string;
    end;

  Tfrm7Zip = class(TForm)
    Label1: TLabel;
    edSource: THistoryCombo;
    Label2: TLabel;
    edFiles: THistoryCombo;
    Label3: TLabel;
    edDest: THistoryCombo;
    btDest: TSpeedButton;
    btFiles: TSpeedButton;
    btSource: TSpeedButton;
    Label4: TLabel;
    cbCompLevel: TComboBox;
    StatusBar: TStatusBar;
    btStart: TBitBtn;
    btQuit: TBitBtn;
    btInfo: TBitBtn;
    edProg: TEdit;
    Label5: TLabel;
    bt7Zip: TSpeedButton;
    cbSubDir: TCheckBox;
    OpenDialog: TOpenDialog;
    cbCrypt: TCheckBox;
    rgComp: TRadioGroup;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure btInfoClick(Sender: TObject);
    procedure btQuitClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure bt7ZipClick(Sender: TObject);
    procedure btStartClick(Sender: TObject);
    procedure btSourceClick(Sender: TObject);
    procedure btFilesClick(Sender: TObject);
    procedure btDestClick(Sender: TObject);
  private
    { Private-Deklarationen }
    ProgVersName,
    ProgVersDate,
    AppPath,UserPath,
    IniName,ProgPath,
    Password            : string;
    Running             : boolean;
    ArcType             : TArcType;
    function Get7Zip (const AName : string) : boolean;
    function CompressFile (const InName,OutName : string) : boolean;
  public
    { Public-Deklarationen }
  end;

var
  frm7Zip: Tfrm7Zip;

implementation

{$R *.dfm}

uses WinUtils, InitProg, FileUtils, StringUtils, ShellDirDlg, IniFiles, GnuGetText,
  LangUtils, NewPwdDlg;

const
  IniExt = 'ini';

  (* INI-Sektionen *)
  CfGSekt = 'Config';
  SourceSekt = 'Source';
  DestSekt = 'Dest';
  FileSekt = 'Files';

  (* INI-Variablen *)
  iniZip   = 'ZipProgram';
  iniSubDir = 'Subdirectories';
  iniArcType = 'Method';
  iniLevel = 'CompressionLevel';

  ArcProps : array [TArcType] of TArcProp = ((TypeId : '-tzip'; Ext : 'zip'),
                                            (TypeId : '-t7z -mhe=on'; Ext : '7z'));

  Levels : array [0..4] of integer=(1,3,5,7,9);

procedure Tfrm7Zip.FormCreate(Sender: TObject);
begin
  TranslateComponent(self);
  InitPaths(AppPath,UserPath,ProgPath);
  InitVersion(_('Compress single files to 7Zip'),'Vers. 1.0',CopRgt,3,3,ProgVersName,ProgVersDate);
  Caption:=_('Compress single files to 7Zip')+' ('+VersInfo.Comments+')';
  IniName:=Erweiter(AppPath,PrgName,IniExt);
  with TIniFile.Create(IniName)do begin
    edProg.Text:=ReadString(CfgSekt,iniZip,'');
    cbSubDir.Checked:=ReadBool(CfgSekt,iniSubDir,false);
    rgComp.ItemIndex:=ReadInteger(CfgSekt,iniArcType,0);
    cbCompLevel.ItemIndex:=ReadInteger(CfgSekt,iniLevel,2);
    Free;
    end;
  with edSource do begin
    LoadFromIni (IniName,SourceSekt);
    ItemIndex:=0;
    end;
  with edDest do begin
    LoadFromIni (IniName,DestSekt);
    ItemIndex:=0;
    end;
  with edFiles do begin
    LoadFromIni (IniName,FileSekt);
    ItemIndex:=0;
    end;
  ArcType:=TArcType(rgComp.ItemIndex);
  Password:='';
  Running:=false;
  end;

procedure Tfrm7Zip.FormDestroy(Sender: TObject);
begin
  with TIniFile.Create(IniName)do begin
    WriteString(CfgSekt,iniZip,edProg.Text);
    WriteBool(CfgSekt,iniSubDir,cbSubDir.Checked);
    WriteInteger(CfgSekt,iniArcType,rgComp.ItemIndex);
    WriteInteger(CfgSekt,iniLevel,cbCompLevel.ItemIndex);
    Free;
    end;
  end;

procedure Tfrm7Zip.FormShow(Sender: TObject);
var
  s : string;
begin
  if (length(edProg.Text)=0) or not FileExists(edProg.Text) then begin
    // search for 7-Zip
    s:=IncludeTrailingPathDelimiter(ProgPath)+Path7Zip+Name7Zip;
    if FileExists(s) then edProg.Text:=s
    else if not Get7Zip(s) then Close;
    end;
  btStart.Enabled:=FileExists(edProg.Text);
  end;

procedure Tfrm7Zip.bt7ZipClick(Sender: TObject);
begin
  Get7Zip(edProg.Text);
  btStart.Enabled:=FileExists(edProg.Text);
  end;

procedure Tfrm7Zip.btInfoClick(Sender: TObject);
begin
  InfoDialog ('',Caption+' - '+ProgVersDate+sLineBreak+CopRgt
           +sLineBreak+'E-Mail: '+EmailAdr);
  end;

function Tfrm7Zip.Get7Zip (const AName : string) : boolean;
begin
  with OpenDialog do begin
    InitialDir:=ExtractFilePath(AName);
    Filename:=ExtractFilename(AName);
    Filter:=_('Programs|*.exe|all|*.*');
    Title:=_('Search for 7Zip program');
    Options:=[ofHideReadOnly,ofPathMustExist,ofFileMustExist,ofEnableSizing];
    Result:=Execute;
    if Result then edProg.Text:=Filename;
    end;
  end;

procedure Tfrm7Zip.btSourceClick(Sender: TObject);
var
  s : string;
begin
  s:=edSource.Text;
  if length(s)=0 then s:=UserPath;
  if ShellDirDialog.Execute (_('Select directory with files to compress'),true,true,true,'',s) then begin
    with edSource do begin
      Text:=s; AddItem(s);
      end;
    with edDest do if length(Text)=0 then begin
      Text:=s; AddItem(s);
      end;
    with edFiles do begin
      Text:='*.*'; AddItem(Text);
      end;
    end;
  end;

procedure Tfrm7Zip.btFilesClick(Sender: TObject);
var
  i : integer;
  s : string;
begin
  with OpenDialog do begin
    InitialDir:=edSource.Text;
    Options:=[ofHideReadOnly,ofAllowMultiSelect,ofPathMustExist,ofFileMustExist,ofEnableSizing];
    Filename:='';
    Title:='Bild-Dateien auswhlen';
    if Execute then begin
      s:=ExtractFilePath(Files.Strings[0]);
      with edDest do if s<>Text then begin
        Text:=s; AddItem(s);
        end;
      with Files do begin
        s:='';
        for i:=0 to Count-1 do s:=s+ExtractFilename(Strings[i])+';';
        end;
      delete (s,length(s),1);
      with edFiles do begin
        Text:=s; AddItem(s);
        end;
      end;
    end;
  end;

procedure Tfrm7Zip.btDestClick(Sender: TObject);
var
  s : string;
begin
  s:=edDest.Text;
  if length(s)=0 then s:=edSource.Text;
  if length(s)=0 then s:=UserPath;
  if ShellDirDialog.Execute(_('Select directory for compressed files'),true,true,true,'',s) then begin
    with edDest do begin
      Text:=s; AddItem(s);
      end;
    end;
  end;

procedure Tfrm7Zip.btQuitClick(Sender: TObject);
begin
  Close;
  end;

function Tfrm7Zip.CompressFile (const InName,OutName : string) : boolean;
var
  si        : TStartupInfo;
  pi        : TProcessInformation;
  ec        : DWord;
  CmdLine   : string;
begin
// Create process to start Program
  FillChar(si, SizeOf(TStartupInfo), 0);
  with si do begin
    cb := Sizeof(TStartupInfo);
    dwFlags := STARTF_USESTDHANDLES or STARTF_USESHOWWINDOW;
    wShowWindow:=SW_HIDE;
    end;
  with ArcProps[ArcType] do begin
    CmdLine:=AnsiQuotedStr(edProg.Text,'"')+' a '
        +AnsiQuotedStr(NewExt(OutName,Ext),'"')+' '+AnsiQuotedStr(InName,'"')
        +Space+TypeId+' -mx'+IntToStr(Levels[cbCompLevel.ItemIndex]);
    end;
  if cbCrypt.Checked then CmdLine:=CmdLine+' -p'+Password;
  if CreateProcess(nil,                // Anwendungsname
                   pchar(CmdLine),
                   nil,                // Security
                   nil,                // Security
                   true,               // use InheritHandles
                   NORMAL_PRIORITY_CLASS, // Prioritt
                   nil,                   // Environment
                   nil,                   // Verzeichnis
                   si,pi) then begin
    while Running and (WaitForSingleObject(pi.hProcess,100)<>WAIT_OBJECT_0) do
      Application.ProcessMessages;
    Result:=GetExitCodeProcess(pi.hProcess,ec);
    CloseHandle(pi.hProcess);
    if not Result then begin // Error
      ErrorDialog(Format(_('Error on executing 7Zip: %s'),[SysErrorMessage(GetLastError)]));
      end;
    end
  else begin
    ErrorDialog(Format(_('Error on starting 7Zip: %s'),[SysErrorMessage(GetLastError)]));
    Result:=false;
    end;
  end;

procedure Tfrm7Zip.btStartClick(Sender: TObject);
var
  FListe     : TStringList;
  s,t        : string;
  i,n        : integer;

  procedure ScanDir (Dir,Mask : string);
  var
    DirInfo    : TSearchRec;
    Findresult : integer;
  begin
    if cbSubDir.Checked then begin
      FindResult:=FindFirst (Erweiter(Dir,'*','*'),faAnyFile,DirInfo);
      while FindResult=0 do with DirInfo do begin
        if NotSpecialDir(Name) then begin
          if((Attr and faDirectory)<>0) then ScanDir (Erweiter(Dir,Name,''),Mask);
          end;
        FindResult:=FindNext (DirInfo);
        end;
      FindClose(DirInfo);
      end;
    FindResult:=FindFirst (Erweiter(Dir,Mask,''),faArchive+faReadOnly+faHidden+faSysfile,DirInfo);
    while FindResult=0 do with DirInfo do begin
      if NotSpecialDir(Name) then FListe.Add(Erweiter(Dir,Name,''));
      FindResult:=FindNext (DirInfo);
      end;
    FindClose(DirInfo);
    end;

begin
  if Running then Running:=false
  else begin
    if cbCrypt.Checked then begin
      if not InputNewPassword(CursorPos,_('Encrypt archive'),Password,true) then Exit;
      end;
    Running:=true;
    btStart.Caption:=_('Stop');
    t:=edFiles.Text;
    n:=0;
    if length(t)>0 then begin
      FListe:=TStringList.Create;
      repeat
        s:=ReadNxtStr (t,';');
        ScanDir (IncludeTrailingPathDelimiter(edSource.Text),s);
        until (length(t)=0) or not Running;
      if Running then with FListe do if Count>0 then for i:=0 to Count-1 do begin
        StatusBar.SimpleText:=_('Compressing: ')+StripPath(Strings[i],50);
        Application.ProcessMessages;
        s:=MakeRelativePath(IncludeTrailingPathDelimiter(edSource.Text),Strings[i]);
        s:=IncludeTrailingPathDelimiter(edDest.Text)+s;
        ForceDirectories(ExtractFilePath(s));
        ArcType:=TArcType(rgComp.ItemIndex);
        if CompressFile(Strings[i],s) then inc(n);
        if not Running then Break;
        end;
      end;
    Running:=false;
    with StatusBar do if n=0 then SimpleText:=_('Nothing to do!')
    else if n=1 then SimpleText:=_('1 file compressed!')
    else SimpleText:=Format(_('%u files compressed!'),[n]);
    btStart.Caption:=_('Start');
    end;
  end;

end.
